﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Shapes;
using Testing.WpfPathTriangulator.Common;
using Testing.WpfPathTriangulator.Graphics;

// Some namespace mappings
using SDX = SharpDX;

namespace Testing.WpfPathTriangulator.ViewModel
{
    class DrawingViewModel : ViewModelBase
    {
        private const double POLYGON_SIMPLIFICATION = 2.0;

        private PathFigure m_pathFigure;
        private PathFigureCollection m_triangulatedPathFigures;

        /// <summary>
        /// Initializes a new instance of the <see cref="DrawingViewModel"/> class.
        /// </summary>
        public DrawingViewModel()
        {
            m_pathFigure = new PathFigure(
                new Point(),
                BuildDefaultPathSegments(),
                true);

            this.TriangulateModelAsync(m_pathFigure)
                .FireAndForeget();
        }

        /// <summary>
        /// Triangulates the given model.
        /// </summary>
        /// <param name="pathGeometry">The path to be triangulated.</param>
        private async Task TriangulateModelAsync(PathFigure pathFigure)
        {
            PathFigure clonedPathFigure = pathFigure.Clone();
            clonedPathFigure.Freeze();

            PathFigureCollection newTrinangulatedPathFigures = null;
            await Task.Factory.StartNew(() =>
            {
                PathGeometry pathGeometry = new PathGeometry(new PathFigure[] { clonedPathFigure });
                var flattened = pathGeometry.GetFlattenedPathGeometry(POLYGON_SIMPLIFICATION, ToleranceType.Absolute);
                foreach(PathFigure actFigure in flattened.Figures)
                {
                    // Build a polygon out of the current figure
                    List<SDX.Vector2> polygonVectors = new List<SDX.Vector2>();
                    polygonVectors.Add(new SDX.Vector2(
                        (float)actFigure.StartPoint.X, (float)actFigure.StartPoint.Y));
                    foreach(PathSegment actSegment in actFigure.Segments)
                    {
                        // Handle line segments
                        LineSegment actLineSegment = actSegment as LineSegment;
                        if(actLineSegment != null)
                        {
                            polygonVectors.Add(new SDX.Vector2(
                                (float)actLineSegment.Point.X, (float)actLineSegment.Point.Y));
                            continue;
                        }

                        // Handle polyline segments
                        PolyLineSegment actPolyLineSegment = actSegment as PolyLineSegment;
                        if(actPolyLineSegment != null)
                        {
                            for(int loop=0 ; loop<actPolyLineSegment.Points.Count; loop++)
                            {
                                polygonVectors.Add(new SDX.Vector2(
                                    (float)actPolyLineSegment.Points[loop].X, (float)actPolyLineSegment.Points[loop].Y));
                            }
                            continue;
                        }
                    }
                    Polygon2D polygon = new Polygon2D(polygonVectors.ToArray());

                    // Triangulate the polygon
                    List<ushort> indices = new List<ushort>(polygon.TriangulateUsingCuttingEars());

                    // Now build new path showing the result of the triangulation
                    if((indices != null) &&
                       (indices.Count > 0))
                    {
                        newTrinangulatedPathFigures = new PathFigureCollection();
                        for(int loop=0 ; loop<indices.Count; loop+= 3)
                        {
                            SDX.Vector2 pointA = polygon.Vertices[indices[loop]];
                            SDX.Vector2 pointB = polygon.Vertices[indices[loop + 1]];
                            SDX.Vector2 pointC = polygon.Vertices[indices[loop + 2]];

                            PathFigure newFigure = new PathFigure();
                            newFigure.StartPoint = new Point((double)pointA.X, (double)pointA.Y);
                            newFigure.Segments.Add(new LineSegment(new Point((double)pointB.X, (double)pointB.Y), true) { IsSmoothJoin = true });
                            newFigure.Segments.Add(new LineSegment(new Point((double)pointC.X, (double)pointC.Y), true) { IsSmoothJoin = true });
                            newFigure.Segments.Add(new LineSegment(new Point((double)pointA.X, (double)pointA.Y), true) { IsSmoothJoin = true });
                            newFigure.IsClosed = false;
                            newFigure.Freeze();
                            newTrinangulatedPathFigures.Add(newFigure);
                        }
                        newTrinangulatedPathFigures.Freeze();
                    }
                }
            });

            if (m_pathFigure == pathFigure)
            {
                this.TriangulatedPathFigures = newTrinangulatedPathFigures;
            }
        }

        /// <summary>
        /// Builds a default path.
        /// </summary>
        private IEnumerable<PathSegment> BuildDefaultPathSegments()
        {
            yield return new LineSegment(new Point(10.0, 10.0), true);
            yield return new LineSegment(new Point(-10.0, 10.0), true);
            yield return new LineSegment(new Point(-10.0, -10.0), true);
            yield return new LineSegment(new Point(10.0, -10.0), true);
            yield break;
        }

        /// <summary>
        /// Gets or sets the currently displayed path.
        /// </summary>
        public PathFigure PathGeometry
        {
            get { return m_pathFigure; }
            set
            {
                if (m_pathFigure != value)
                {
                    m_pathFigure = value;
                    RaisePropertyChanged(() => PathGeometry);
                    RaisePropertyChanged(() => GeneratedFigures);

                    this.m_triangulatedPathFigures = null;
                    TriangulateModelAsync(m_pathFigure)
                        .FireAndForeget();
                }
            }
        }

        /// <summary>
        /// Gets or sets the path figure for the triangulated model.
        /// </summary>
        public PathFigureCollection TriangulatedPathFigures
        {
            get { return m_triangulatedPathFigures; }
            set
            {
                if(m_triangulatedPathFigures != value)
                {
                    m_triangulatedPathFigures = value;
                    RaisePropertyChanged(() => TriangulatedPathFigures);
                }
            }
        }

        /// <summary>
        /// Gets the generated figureds (for binding with PathGeometry object).
        /// </summary>
        public PathFigureCollection GeneratedFigures
        {
            get
            {
                if (m_pathFigure == null) { return new PathFigureCollection(); }

                return new PathFigureCollection(new PathFigure[] { m_pathFigure }); 
            }
        }

        /// <summary>
        /// Gets the generated simplified figures (for binding with PathGeometry object).
        /// </summary>
        public PathFigureCollection GeneratedSimplifiedFigures
        {
            get
            {
                if (m_pathFigure == null) { return new PathFigureCollection(); }

                PathGeometry pathGeometry = new PathGeometry(new PathFigure[] { m_pathFigure });
                var flattened = pathGeometry.GetFlattenedPathGeometry(POLYGON_SIMPLIFICATION, ToleranceType.Absolute);
                return flattened.Figures;
            }
        }
    }
}
